//+------------------------------------------------------------------+
//|                                                ParticleSwarm.mqh |
//|                                      Copyright  2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes:                                                 |
//|  Particle Swarm Speeds Up Expert Adviser Optimization On The Fly |
//|                        https://www.mql5.com/en/blogs/post/683577 |
//+------------------------------------------------------------------+

#ifndef _PSO_H_
#define _PSO_H_

#property strict

#include <TypeToBytes.mqh>
#include <crc64.mqh>

template<typename T>
ulong crc64(const T &array[])
{
  ulong crc = 0;
  int len = ArraySize(array);
  for(int i = 0; i < len; i++)
  {
    crc = crc64(crc, _R(array[i]).Bytes, sizeof(T));
  }
  return crc;
}

template<typename T>
class TreeNode
{
  private:
    TreeNode *left;
    TreeNode *right;
    T value;

  public:
    TreeNode(T t)
    {
      value = t;
    }
    
    bool Add(T t)
    {
      if(t < value)
      {
        if(left == NULL)
        {
          left = new TreeNode<T>(t);
          return false;
        }
        else
        {
          return left.Add(t);
        }
      }
      else if(t > value)
      {
        if(right == NULL)
        {
          right = new TreeNode<T>(t);
          return false;
        }
        else
        {
          return right.Add(t);
        }
      }
      return true;
    }
    
    ~TreeNode()
    {
      if(left != NULL) delete left;
      if(right != NULL) delete right;
    }
};

template<typename T>
class BinaryTree
{
  private:
    TreeNode<T> *root;
    
  public:
    bool Add(T t)
    {
      if(root == NULL)
      {
        root = new TreeNode<T>(t);
        return false;
      }
      else
      {
        return root.Add(t);
      }
    }
    
    ~BinaryTree()
    {
      if(root != NULL) delete root;
    }
};

class Particle
{
  public:
    double position[];
    double best[];
    double velocity[];
    
    double positionValue;
    double bestValue;
    
    Particle(const int params)
    {
      ArrayResize(position, params);
      ArrayResize(best, params);
      ArrayResize(velocity, params);
      bestValue = -DBL_MAX;
    }
};

class Group
{
  private:
    double result;
  
  public:
    double optimum[];
    
    Group(const int params)
    {
      ArrayResize(optimum, params);
      ArrayInitialize(optimum, 0);
      result = -DBL_MAX;
    }
    
    void assign(const double x)
    {
      result = x;
    }
    
    double getResult() const
    {
      return result;
    }
    
    bool isAssigned()
    {
      return result != -DBL_MAX;
    }
};

class Functor
{
  public:
    virtual double calculate(const double &vector[]) = 0;
};

class Swarm
{
  private:
    Particle *particles[];
    Group *groups[];
    int _size;
    int _globals;
    int _params;
    
    double highs[];
    double lows[];
    double steps[];
    double solution[];
    
    BinaryTree<ulong> index;
    
    static int particle2group[];
    
    void init(const int size, const int globals, const int params, const double &max[], const double &min[], const double &step[])
    {
      // TODO: check that product of all ((max - min)/step) is much less than size
      // TODO: check that globals is much less than size
      ArrayResize(particles, size);
      for(int i = 0; i < size; i++)
      {
        particles[i] = new Particle(params);
      }
      _size = size;
      
      ArrayResize(groups, globals);
      for(int i = 0; i < globals; i++)
      {
        groups[i] = new Group(params);
      }
      _globals = globals;
      
      ArrayResize(particle2group, size);
      
      for(int i = 0; i < size; i++)
      {
        particle2group[i] = (int)MathMin(MathRand() * 1.0 / 32767 * globals, globals - 1);
      }
      
      ArrayResize(highs, params);
      ArrayResize(lows, params);
      ArrayResize(steps, params);
      for(int i = 0; i < params; i++)
      {
        highs[i] = max[i];
        lows[i] = min[i];
        steps[i] = step[i];
      }
      
      ArrayResize(solution, params);
      _params = params;
      Print("PSO[", params, "] created: ", size, "/", globals);
    }
    
    void arrayPrint(const string title, const double &V[], const int accuracy)
    {
      int n = ArraySize(V);
      string s;
      for(int i = 0; i < n; i++)
      {
        s = s + " " + DoubleToString(V[i], accuracy);
      }
      Print(title + ":" + s);
    }    
    
  public:
    Swarm(const int params, const double &max[], const double &min[], const double &step[])
    {
      init(params * 5, (int)MathSqrt(params * 5), params, max, min, step);
    }
    
    Swarm(const int size, const int globals, const int params, const double &max[], const double &min[], const double &step[])
    {
      init(size, globals, params, max, min, step);
    }
    
    ~Swarm()
    {
      for(int i = 0; i < _size; i++)
      {
        delete particles[i];
      }
      for(int i = 0; i < _globals; i++)
      {
        delete groups[i];
      }
    }
    
    int getSize()
    {
      return _size;
    }
    
    double optimize(Functor &f, const int cycles, const double inertia = 0.8, const double selfBoost = 0.4, const double groupBoost = 0.4)
    {
      double result = -DBL_MAX;
      ArrayInitialize(solution, 0);
      
      for(int i = 0; i < _size; i++)
      {
        do
        {
          for(int p = 0; p < _params; p++)
          {
            particles[i].position[p] = (MathRand() * 1.0 / 32767) * (highs[p] - lows[p]) + lows[p];
            if(steps[p] != 0)
            {
              particles[i].position[p] = ((int)MathRound(particles[i].position[p] / steps[p])) * steps[p];
            }
            particles[i].best[p] = particles[i].position[p];
            particles[i].velocity[p] = (MathRand() * 1.0 / 32767) * 2 * (highs[p] - lows[p]) - (highs[p] - lows[p]);
          }
        }
        while(index.Add(crc64(particles[i].position)) && !IsStopped());
        
        particles[i].positionValue = f.calculate(particles[i].position);
        if(!groups[particle2group[i]].isAssigned() || particles[i].positionValue > groups[particle2group[i]].getResult())
        {
          groups[particle2group[i]].assign(particles[i].positionValue);
          for(int p = 0; p < _params; p++)
          {
            groups[particle2group[i]].optimum[p] = particles[i].position[p];
          }
          
          if(particles[i].positionValue > result)
          {
            result = particles[i].positionValue;
            for(int p = 0; p < _params; p++)
            {
              solution[p] = particles[i].position[p];
            }
          }
        }
      }
      
      double next[];
      ArrayResize(next, _params);
      
      int actualCount = 0;
    
      Print("PSO Processing...");
      for(int c = 0; c < cycles && !IsStopped(); c++)
      {
        int skipped = 0;
        for(int i = 0; i < _size && !IsStopped(); i++)
        {
          bool alreadyProcessed;
          for(int p = 0; p < _params; p++)
          {
            double r1 = MathRand() * 1.0 / 32767;
            double rg = MathRand() * 1.0 / 32767;
            particles[i].velocity[p] = inertia * particles[i].velocity[p] + selfBoost * r1 * (particles[i].best[p] - particles[i].position[p]) + groupBoost * rg * (groups[particle2group[i]].optimum[p] - particles[i].position[p]);
            next[p] = particles[i].position[p] + particles[i].velocity[p];
            if(next[p] < lows[p]) next[p] = lows[p];
            else if(next[p] > highs[p]) next[p] = highs[p];
            if(steps[p] != 0)
            {
              next[p] = ((int)MathRound(next[p] / steps[p])) * steps[p];
            }
          }

          alreadyProcessed = index.Add(crc64(next));
          
          if(alreadyProcessed)
          {
            skipped++;
            continue;
          }
          
          actualCount++;
          
          for(int p = 0; p < _params; p++)
          {
            particles[i].position[p] = next[p];
          }
          
          particles[i].positionValue = f.calculate(particles[i].position);
          // arrayPrint((string)i + " " + (string)particles[i].positionValue, next, 3);
          if(particles[i].positionValue > particles[i].bestValue)
          {
            particles[i].bestValue = particles[i].positionValue;
            for(int p = 0; p < _params; p++)
            {
              particles[i].best[p] = particles[i].position[p];
            }
          }
    
          if(particles[i].positionValue > groups[particle2group[i]].getResult())
          {
            groups[particle2group[i]].assign(particles[i].positionValue);
            for(int p = 0; p < _params; p++)
            {
              groups[particle2group[i]].optimum[p] = particles[i].position[p];
            }
            
            if(particles[i].positionValue > result)
            {
              result = particles[i].positionValue;
              for(int p = 0; p < _params; p++)
              {
                solution[p] = particles[i].position[p];
              }
            }
          }
        }
        Print("Pass ", c, " done, skipped ", skipped, " of ", _size, " / ", result);
      }
      Print("PSO Finished ", actualCount, " of ", (cycles * _size), " planned passes: ", !IsStopped());
      return result;
    }
    
    bool getSolution(double &result[])
    {
      ArrayResize(result, _params);
      for(int p = 0; p < _params; p++)
      {
        result[p] = solution[p];
      }
      return !IsStopped();
    }
    
};

static int Swarm::particle2group[];

#endif